#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import importlib.util
import inspect
import io
import os
import re
import shutil
import ssl
import subprocess
import sys
import tarfile
import time
import urllib.request
import zipfile
from itertools import chain
from types import ModuleType, FunctionType
from typing import Dict, List, Tuple, Callable, Union, no_type_check, Set, Any, cast, TextIO, Optional

import yaml

from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine import jinja, thirdparties
from kubemarine.core import utils, static, errors, os as kos, log
from kubemarine.core.yaml_merger import default_merger
from kubemarine.core.group import NodeGroup
from kubemarine.kubernetes.daemonset import DaemonSet
from kubemarine.kubernetes.deployment import Deployment
from kubemarine.kubernetes.replicaset import ReplicaSet
from kubemarine.kubernetes.statefulset import StatefulSet

# list of plugins owned and managed by kubemarine
oob_plugins = list(static.DEFAULTS["plugins"].keys())
LOADED_MODULES: Dict[str, ModuleType] = {}

ERROR_PODS_NOT_READY = 'In the expected time, the pods did not become ready'


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    for plugin_name, plugin_item in cluster.inventory["plugins"].items():
        for step in plugin_item.get('installation', {}).get('procedures', []):
            for procedure_type, configs in step.items():
                if procedure_types()[procedure_type].get('verify') is not None:
                    procedure_types()[procedure_type]['verify'](cluster, configs, plugin_name)


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    # it is necessary to convert URIs from quay.io/xxx:v1 to example.com:XXXX/xxx:v1
    plugins_default_registry = inventory['plugin_defaults']['installation'].get('registry')

    if plugins_default_registry is not None:
        # The following section rewrites DEFAULT plugins registries
        # and does not touch user-defined registries in plugins
        for plugin_name, plugin_item in cluster.inventory['plugins'].items():
            if cluster.raw_inventory.get('plugins', {}).get(plugin_name, {}).get('installation', {}).get('registry') is None:
                plugin_item.setdefault('installation', {})['registry'] = plugins_default_registry

    for plugin_name, plugin_item in inventory["plugins"].items():
        for step in plugin_item.get('installation', {}).get('procedures', []):
            for procedure_type, configs in step.items():
                if procedure_types()[procedure_type].get('convert') is not None:
                    step[procedure_type] = procedure_types()[procedure_type]['convert'](cluster, configs)


def _get_upgrade_plan(cluster: KubernetesCluster) -> List[Tuple[str, dict]]:
    context = cluster.context
    procedure = context["initial_procedure"]
    procedure_inventory = cluster.procedure_inventory
    upgrade_plan = []
    if procedure == "upgrade":
        kubernetes_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        for i, v in enumerate(procedure_inventory['upgrade_plan'][context['upgrade_step']:]):
            # Take the target (probably) compiled version and the remained not yet compiled
            version = kubernetes_version if i == 0 else v
            upgrade_plan.append((version, procedure_inventory.get(v, {}).get("plugins", {})))

    elif 'upgrading_plugin' in context:
        upgrade_plugins = procedure_inventory.get('upgrade', {}).get("plugins", {})
        upgrade_plugins = utils.subdict_yaml(upgrade_plugins, [context['upgrading_plugin']])
        upgrade_plan = [("", upgrade_plugins)]

    return upgrade_plan


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine'])
def enrich_upgrade_inventory(cluster: KubernetesCluster) -> None:
    upgrade_plan = _get_upgrade_plan(cluster)
    upgrade_plugins = {} if not upgrade_plan else upgrade_plan[0][1]

    if upgrade_plugins:
        default_merger.merge(cluster.inventory.setdefault("plugins", {}), utils.deepcopy_yaml(upgrade_plugins))


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine'])
def verify_upgrade_inventory(cluster: KubernetesCluster) -> None:
    upgrade_plan = _get_upgrade_plan(cluster)
    if not upgrade_plan:
        return

    context = cluster.context
    if context.get("initial_procedure") == "upgrade":
        previous_version = cluster.previous_inventory['services']['kubeadm']['kubernetesVersion']
        plugins_verify = oob_plugins
    else:  # migrate_kubemarine procedure
        previous_version = ""
        plugins_verify = [context['upgrading_plugin']]

    _verify_upgrade_plan(cluster, previous_version, plugins_verify, upgrade_plan)


def _verify_upgrade_plan(cluster: KubernetesCluster, previous_version: str,
                         plugins_verify: List[str], upgrade_plan: List[Tuple[str, dict]]) -> None:
    raw_plugins = utils.deepcopy_yaml(cluster.previous_raw_inventory.get('plugins', {}))

    # validate all plugin sections in procedure inventory
    for version, upgrade_plugins in upgrade_plan:
        if version != '' and jinja.is_template(version):
            break

        for plugin_name in plugins_verify:
            verify_image_redefined(plugin_name,
                                   previous_version,
                                   version,
                                   raw_plugins.get(plugin_name, {}),
                                   upgrade_plugins.get(plugin_name, {}))
        default_merger.merge(raw_plugins, utils.deepcopy_yaml(upgrade_plugins))
        previous_version = version


def verify_image_redefined(plugin_name: str, previous_version: str, next_version: str,
                           raw_plugins: Dict[str, Any], upgrade_plugin: Dict[str, Any]) -> None:
    """
    If some image in "cluster_plugin" is different from image in "base_plugin",
    i.e. redefined, then "upgrade_plugin" should have this image explicitly
    redefined too.
    """
    sensitive_keys = ['image', 'helper-pod-image', 'version']
    for key, value in raw_plugins.items():
        if isinstance(value, dict):
            verify_image_redefined(plugin_name,
                                   previous_version,
                                   next_version,
                                   value,
                                   upgrade_plugin.get(key, {}))
        elif key not in sensitive_keys:
            continue
        elif value and not upgrade_plugin.get(key):
            raise errors.KME("KME0009",
                             key=key, plugin_name=plugin_name,
                             previous_version_spec=f" for version {previous_version}" if previous_version else "",
                             next_version_spec=f" for next version {next_version}" if next_version else ""
            )


def _get_plugin_priority(plugin_item: dict, default: int) -> int:
    priority: int = plugin_item.get("installation", {}).get('priority', default)
    return priority


def install(cluster: KubernetesCluster, plugins_: Dict[str, dict] = None) -> None:
    if plugins_ is None:
        plugins = cluster.inventory["plugins"]
    else:
        plugins = plugins_
    plugins_queue: List[str] = []
    max_priority = 0
    for plugin_name, plugin_item in plugins.items():
        if plugin_item.get("install", False) and plugin_item.get("installation", {}).get('procedures') is not None:
            plugins_queue.append(plugin_name)
            if plugin_item.get("installation", {}).get('priority') is not None \
                    and plugin_item['installation']['priority'] > max_priority:
                max_priority = plugin_item['installation']['priority']

    plugins_queue.sort(key=lambda name: _get_plugin_priority(plugins[name], max_priority + 1))

    cluster.log.debug('The following plugins will be installed:')
    for plugin_name in plugins_queue:
        cluster.log.debug('%i. %s' % (
            plugins[plugin_name].get("installation", {}).get('priority', max_priority + 1),
            plugin_name
        ))

    cluster.log.debug('Starting plugins installation:')

    for plugin_name in plugins_queue:
        install_plugin(cluster, plugin_name, plugins[plugin_name]["installation"]['procedures'])


def install_plugin(cluster: KubernetesCluster, plugin_name: str, installation_procedure: List[dict]) -> None:
    cluster.log.debug("**** INSTALLING PLUGIN %s ****" % plugin_name)

    for step in installation_procedure:
        for apply_type, configs in step.items():
            procedure_types()[apply_type]['apply'](cluster, configs)


def expect_daemonset(cluster: KubernetesCluster,
                     daemonsets_names: List[Union[str, Dict[str, str]]],
                     timeout: int = None,
                     retries: int = None,
                     node: NodeGroup = None) -> None:
    """
    The method waits for the configuration parameters of the given DaemonSets to be applied.
    :param cluster: KubernetesCluster object where method should be performed
    :param daemonsets_names: List of DaemonSet names (or dicts with name and namespace) to be
    expected
    :param timeout: Retry attempt time (seconds)
    :param retries: Number of retry attempts
    :param node: Node where daemonsets should be detected
    :return: None
    """

    log = cluster.log

    if timeout is None:
        timeout = cluster.inventory['globals']['expect']['deployments']['timeout']
    if retries is None:
        retries = cluster.inventory['globals']['expect']['deployments']['retries']

    log.debug(f"Expecting the following DaemonSets to be up to date: {daemonsets_names}")
    log.verbose("Max expectation time: %ss" % (timeout * retries))

    log.debug("Waiting for DaemonSets...")

    daemonsets = []
    for name in daemonsets_names:
        if isinstance(name, str):
            daemonsets.append(DaemonSet(cluster, name=name, namespace='kube-system'))
        elif isinstance(name, dict):
            daemonsets.append(DaemonSet(cluster, name=name['name'], namespace=name['namespace']))

    while retries > 0:
        up_to_date = True
        for daemonset in daemonsets:
            if not daemonset.reload(control_plane=node, suppress_exceptions=True).is_up_to_date():
                up_to_date = False

        if up_to_date:
            cluster.log.debug("DaemonSets are up to date")
            return
        else:
            retries -= 1
            cluster.log.debug(f"DaemonSets are not up to date yet... ({retries * timeout}s left)")
            time.sleep(timeout)

    raise Exception('In the expected time, the DaemonSets did not become ready. '
                    'Try to increase number of retries in expect.daemonsets: '
                    # pylint: disable-next=line-too-long
                    'https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#expect-deploymentsdaemonsetsreplicasetsstatefulsets')


def expect_replicaset(cluster: KubernetesCluster,
                      replicasets_names: List[Union[str, Dict[str, str]]],
                      timeout: int = None,
                      retries: int = None,
                      node: NodeGroup = None) -> None:
    """
    The method waits for the configuration parameters of the given ReplicaSets to be applied.
    :param cluster: KubernetesCluster object where method should be performed
    :param replicasets_names: List of ReplicaSets names (or dicts with name and namespace) to be
    expected
    :param timeout: Retry attempt time (seconds)
    :param retries: Number of retry attempts
    :param node: Node where replicasests should be detected
    :return: None
    """

    log = cluster.log

    if timeout is None:
        timeout = cluster.inventory['globals']['expect']['deployments']['timeout']
    if retries is None:
        retries = cluster.inventory['globals']['expect']['deployments']['retries']

    log.debug(f"Expecting the following ReplicaSets to be up to date: {replicasets_names}")
    log.verbose("Max expectation time: %ss" % (timeout * retries))

    log.debug("Waiting for ReplicaSets...")

    replicasets = []
    for name in replicasets_names:
        if isinstance(name, str):
            replicasets.append(ReplicaSet(cluster, name=name, namespace='kube-system'))
        elif isinstance(name, dict):
            replicasets.append(ReplicaSet(cluster, name=name['name'], namespace=name['namespace']))

    while retries > 0:
        up_to_date = True
        for replicaset in replicasets:
            if not replicaset.reload(control_plane=node, suppress_exceptions=True).is_available():
                up_to_date = False

        if up_to_date:
            cluster.log.debug("ReplicaSets are up to date")
            return
        else:
            retries -= 1
            cluster.log.debug(f"ReplicaSets are not up to date yet... ({retries * timeout}s left)")
            time.sleep(timeout)

    raise Exception('In the expected time, the ReplicaSets did not become ready. '
                    'Try to increase number of retries in expect.replicasets: '
                    # pylint: disable-next=line-too-long
                    'https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#expect-deploymentsdaemonsetsreplicasetsstatefulsets')


def expect_statefulset(cluster: KubernetesCluster,
                       statefulsets_names: List[Union[str, Dict[str, str]]],
                       timeout: int = None,
                       retries: int = None,
                       node: NodeGroup = None) -> None:
    """
    The method waits for the configuration parameters of the given StatefulSets to be applied.
    :param cluster: KubernetesCluster object where method should be performed
    :param statefulsets_names: List of StatefulSets names (or dicts with name and namespace) to be
    expected
    :param timeout: Retry attempt time (seconds)
    :param retries: Number of retry attempts
    :param node: Node where statefulsets should be detected
    :return: None
    """

    log = cluster.log

    if timeout is None:
        timeout = cluster.inventory['globals']['expect']['deployments']['timeout']
    if retries is None:
        retries = cluster.inventory['globals']['expect']['deployments']['retries']

    log.debug(f"Expecting the following StatefulSets to be up to date: {statefulsets_names}")
    log.verbose("Max expectation time: %ss" % (timeout * retries))

    log.debug("Waiting for StatefulSets...")

    statefulsets = []
    for name in statefulsets_names:
        if isinstance(name, str):
            statefulsets.append(StatefulSet(cluster, name=name, namespace='kube-system'))
        elif isinstance(name, dict):
            statefulsets.append(StatefulSet(cluster, name=name['name'], namespace=name['namespace']))

    while retries > 0:
        up_to_date = True
        for statefulset in statefulsets:
            if not statefulset.reload(control_plane=node, suppress_exceptions=True).is_updated():
                up_to_date = False

        if up_to_date:
            cluster.log.debug("StatefulSets are up to date")
            return
        else:
            retries -= 1
            cluster.log.debug(f"StatefulSets are not up to date yet... ({retries * timeout}s left)")
            time.sleep(timeout)

    raise Exception('In the expected time, the StatefulSets did not become ready. '
                    'Try to increase number of retries in expect.statefulsets: '
                    # pylint: disable-next=line-too-long
                    'https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#expect-deploymentsdaemonsetsreplicasetsstatefulsets')


def expect_deployment(cluster: KubernetesCluster,
                      deployments_names: List[Union[str, Dict[str, str]]],
                      timeout: int = None,
                      retries: int = None,
                      node: NodeGroup = None) -> None:
    """
    The method waits for the configuration parameters of the given Deployments to be applied.
    :param cluster: KubernetesCluster object where method should be performed
    :param deployments_names: List of Deployments names (or dicts with name and namespace) to be
    expected
    :param timeout: Retry attempt time (seconds)
    :param retries: Number of retry attempts
    :param node: Node where deployments should be detected
    :return: None
    """

    log = cluster.log

    if timeout is None:
        timeout = cluster.inventory['globals']['expect']['deployments']['timeout']
    if retries is None:
        retries = cluster.inventory['globals']['expect']['deployments']['retries']

    log.debug(f"Expecting the following Deployments to be up to date: {deployments_names}")
    log.verbose("Max expectation time: %ss" % (timeout * retries))

    log.debug("Waiting for Deployments...")

    deployments = []
    for name in deployments_names:
        if isinstance(name, str):
            deployments.append(Deployment(cluster, name=name, namespace='kube-system'))
        elif isinstance(name, dict):
            deployments.append(Deployment(cluster, name=name['name'], namespace=name['namespace']))

    while retries > 0:
        up_to_date = True
        for deployment in deployments:
            if not deployment.reload(control_plane=node, suppress_exceptions=True).is_actual_and_ready():
                up_to_date = False

        if up_to_date:
            cluster.log.debug("Deployments are up to date!")
            return
        else:
            retries -= 1
            cluster.log.debug(f"Deployments are not up to date yet... ({retries * timeout}s left)")
            time.sleep(timeout)

    raise Exception('In the expected time, the Deployments did not become ready. '
                    'Try to increase number of retries in expect.deployments: '
                    # pylint: disable-next=line-too-long
                    'https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#expect-deploymentsdaemonsetsreplicasetsstatefulsets')


def expect_pods(cluster: KubernetesCluster, pods: List[str], namespace: str = None,
                timeout: int = None, retries: int = None,
                control_plane: NodeGroup = None, node_name: str = None) -> None:

    if timeout is None:
        timeout = cluster.inventory['globals']['expect']['pods']['plugins']['timeout']
    if retries is None:
        retries = cluster.inventory['globals']['expect']['pods']['plugins']['retries']

    cluster.log.debug("Expecting the following pods to be ready: %s" % pods)
    cluster.log.verbose("Max expectation time: %ss" % (timeout * retries))

    cluster.log.debug("Waiting for pods...")

    failures = 0

    if control_plane is None:
        control_plane = cluster.nodes['control-plane'].get_first_member()

    namespace_filter = '-A'
    if namespace is not None:
        namespace_filter = "-n " + namespace

    command = f"kubectl get pods {namespace_filter} -o=wide"
    if node_name is not None:
        command += ' | grep %s' % node_name

    while retries > 0:

        result = control_plane.sudo(command, warn=True, pty=True)

        stdout = list(result.values())[0].stdout
        running_pods_stdout = ''

        failure_found = False

        for stdout_line in iter(stdout.splitlines()):

            # is current line has requested pod for verification?
            # we do not have to fail on pods with bad status which was not requested
            for pod in pods:
                if pod + "-" not in stdout_line:
                    continue

                if is_critical_state_in_stdout(cluster, stdout_line):
                    cluster.log.verbose("Failed pod detected: %s\n" % stdout_line)

                    if not failure_found:
                        failure_found = True
                        failures += 1

                    # just in case, skip the error a couple of times, what if it comes out of the failure state?
                    if failures > cluster.globals['pods']['allowed_failures']:
                        raise Exception('Pod entered a state of error, further proceeding is impossible')
                else:
                    # we have to take into account any pod in not a critical state
                    running_pods_stdout += stdout_line + '\n'

                break

        pods_ready = False
        if running_pods_stdout and running_pods_stdout != "" and "0/1" not in running_pods_stdout:
            pods_ready = True
            for pod in pods:
                # it is necessary to look for pods with the name "xxxx-xxxx-" instead of "xxxx-xxxx" because
                # "xxxx-xxxx" may be the name of the namespace in which another healthy pod will be running
                if pod + "-" not in running_pods_stdout:
                    pods_ready = False

        if pods_ready:
            cluster.log.debug("Pods are ready!")
            cluster.log.debug(running_pods_stdout)
            return
        else:
            retries -= 1
            cluster.log.debug("Pods are not ready yet... (%ss left)" % (retries * timeout))
            cluster.log.debug(running_pods_stdout)
            time.sleep(timeout)

    raise Exception(ERROR_PODS_NOT_READY)


def is_critical_state_in_stdout(cluster: KubernetesCluster, stdout: str) -> bool:
    for state in cluster.globals['pods']['critical_states']:
        if state in stdout:
            return True
    return False


# **** TEMPLATES ****

def convert_template(_: KubernetesCluster, config: Union[str, dict]) -> dict:
    return _convert_file(config)


def verify_template(_: KubernetesCluster, config: dict, _plugin_name: Optional[str] = None) -> None:
    _verify_file(config, "Template")


def apply_template(cluster: KubernetesCluster, config: dict) -> None:
    _apply_file(cluster, config, "Template")


# **** EXPECT ****

def convert_expect(_: KubernetesCluster, config: dict) -> dict:
    if config.get('daemonsets') is not None and isinstance(config['daemonsets'], list):
        config['daemonsets'] = {
            'list': config['daemonsets']
        }
    if config.get('replicasets') is not None and isinstance(config['replicasets'], list):
        config['replicasets'] = {
            'list': config['replicasets']
        }
    if config.get('statefulsets') is not None and isinstance(config['statefulsets'], list):
        config['statefulsets'] = {
            'list': config['statefulsets']
        }
    if config.get('deployments') is not None and isinstance(config['deployments'], list):
        config['deployments'] = {
            'list': config['deployments']
        }
    if config.get('pods') is not None and isinstance(config['pods'], list):
        config['pods'] = {
            'list': config['pods']
        }
    return config


def apply_expect(cluster: KubernetesCluster, config: dict) -> None:
    # TODO: Add support for expect services and expect nodes

    for expect_type in config:
        if expect_type == 'daemonsets':
            expect_daemonset(cluster, config['daemonsets']['list'],
                             timeout=config['daemonsets'].get('timeout'),
                             retries=config['daemonsets'].get('retries'))

        elif expect_type == 'replicasets':
            expect_replicaset(cluster, config['replicasets']['list'],
                              timeout=config['replicasets'].get('timeout'),
                              retries=config['replicasets'].get('retries'))

        elif expect_type == 'statefulsets':
            expect_statefulset(cluster, config['statefulsets']['list'],
                               timeout=config['statefulsets'].get('timeout'),
                               retries=config['statefulsets'].get('retries'))

        elif expect_type == 'deployments':
            expect_deployment(cluster, config['deployments']['list'],
                              timeout=config['deployments'].get('timeout'),
                              retries=config['deployments'].get('retries'))

        elif expect_type == 'pods':
            expect_pods(cluster, config['pods']['list'], namespace=config['pods'].get('namespace'),
                        timeout=config['pods'].get('timeout'),
                        retries=config['pods'].get('retries'))

# **** PYTHON ****


@no_type_check
def get_python_module(module_path: str) -> ModuleType:
    if module_path in LOADED_MODULES:
        return LOADED_MODULES[module_path]

    spec = importlib.util.spec_from_file_location('module', module_path)
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        LOADED_MODULES[module_path] = module
    except Exception as e:
        raise ValueError(f"Could not import module {module_path}: {e}") from None
    return module 


def get_python_method_args(step: dict) -> Tuple[str, FunctionType, Dict[str, Any]]:
    module_path, _ = utils.determine_resource_absolute_file(step['module'])
    method_name = step['method']
    method_arguments = step.get('arguments', {})

    module = get_python_module(module_path)

    # Check if the method exists
    if not hasattr(module, method_name):
        raise ValueError(f"Module {module_path} does not have method {method_name}")

    # Get the method object
    method = getattr(module, method_name)

    return module_path, method, method_arguments


def verify_python(cluster: KubernetesCluster, step: dict, plugin_name: Optional[str] = None) -> None:
    _, method, method_arguments = get_python_method_args(step)

    # Get the signature of the method
    signature = inspect.signature(method)

    # Check if the passed arguments match the signature
    try:
        signature.bind(cluster, **method_arguments)
    except TypeError as e:
        raise ValueError(f"Invalid arguments for python method {method.__name__} for {plugin_name!r} plugin: {e}") \
            from None


def apply_python(cluster: KubernetesCluster, step: dict) -> None:
    module_path, method, method_arguments = get_python_method_args(step)
    cluster.log.debug("Running method %s from %s module..." % (method.__name__, module_path))
    method(cluster, **method_arguments)


# **** THIRDPARTIES ****

def verify_thirdparty(cluster: KubernetesCluster, thirdparty: str, _plugin_name: Optional[str] = None) -> None:
    defined_thirdparties = list(cluster.inventory['services'].get('thirdparties', {}).keys())
    if thirdparty not in defined_thirdparties:
        raise Exception('Specified thirdparty %s not found in thirdpartirs definition. Expected any of %s.'
                        % (thirdparty, defined_thirdparties))


def apply_thirdparty(cluster: KubernetesCluster, thirdparty: str) -> None:
    thirdparties.install_thirdparty(cluster.nodes['all'], thirdparty)


# **** SHELL ****

def convert_shell(_: KubernetesCluster, config: Union[str, dict]) -> dict:
    if isinstance(config, str):
        config = {
            'command': config
        }
    return config


def verify_shell(cluster: KubernetesCluster, config: dict, _plugin_name: Optional[str] = None) -> None:
    out_vars = config.get('out_vars', [])
    groups = config.get('groups', [])
    nodes = config.get('nodes', [])
    explicit_group = cluster.create_group_from_groups_nodes_names(groups, nodes)
    if out_vars and (groups or nodes) and explicit_group.nodes_amount() != 1:
        raise Exception('Shell output variables could be used for single-node groups, but multi-node group was found')

    in_vars = config.get('in_vars', [])
    words_splitter = re.compile(r'\W')
    for var in chain(in_vars, out_vars):
        var_name = var['name']
        if len(words_splitter.split(var_name)) > 1:
            raise Exception(f"'{var_name}' is not a valid shell variable name")

    # TODO: verify fields types and contents


def apply_shell(cluster: KubernetesCluster, step: dict) -> None:
    commands = step['command']
    sudo = step.get('sudo', False)
    groups = step.get('groups', [])
    nodes = step.get('nodes', [])
    in_vars = step.get('in_vars', [])
    out_vars = step.get('out_vars', [])
    vars_separator = "~~~~EXPORTED_VARIABLE~~~~"

    if not groups and not nodes:
        common_group = cluster.nodes['control-plane'].get_any_member()
    else:
        common_group = cluster.create_group_from_groups_nodes_names(groups, nodes)

    if isinstance(commands, list):
        commands = ' && '.join(commands)

    out_vars_aliases: Dict[str, Set[str]] = {}
    for var in out_vars:
        var_name = var['name']
        if var_name in out_vars_aliases:
            # var is already exported, need to only add alternative alias
            out_vars_aliases[var_name].add(var.get('save_as', var_name))
            continue

        out_vars_aliases[var_name] = {var.get('save_as', var_name)}
        # print variable info to stdout in yaml format, separating data using `vars_separator`
        # quotes usage is important for following code to work correctly in different cases
        echo_var_cmd = f"echo {vars_separator} && " \
            f"echo name: {var_name} && " \
            f"echo 'value: |2-' && " \
            f"echo \"${var_name}\" | sed 's/^/  /'"
        commands = f"{commands} && {echo_var_cmd}"

    in_vars_dict: Dict[str, str] = {}
    for var in in_vars:
        var_name = var['name']
        # get defined value or saved value, defaulting to empty value
        var_value = var.get('value', cluster.context['runtime_vars'].get(var_name, ''))
        # replace single-quotes with '"'"' to isolate all single quotes during ssh env inlining
        var_value = var_value.replace("'", "'\"'\"'")
        # wrap variable value with single-quotes for `inline_ssh_env` feature to work correctly with different content
        in_vars_dict[var_name] = f"'{var_value}'"

    cluster.log.debug('Running shell command...')
    if sudo:
        result = common_group.sudo(commands, env=in_vars_dict)
    else:
        result = common_group.run(commands, env=in_vars_dict)

    if out_vars:
        stdout = list(result.values())[0].stdout
        stdout_parts = stdout.split(vars_separator)
        cluster.log.debug(stdout_parts[0])  # printing original user output
        for part in stdout_parts[1:]:
            var = yaml.safe_load(part)
            aliases = out_vars_aliases[var['name']]
            for alias in aliases:
                cluster.context['runtime_vars'][alias] = var['value']
    else:
        cluster.log.debug(result)


# **** ANSIBLE ****

def convert_ansible(_: KubernetesCluster, config: Union[str, dict]) -> dict:
    if isinstance(config, str):
        config = {
            'playbook': config
        }
    return config


def _get_absolute_playbook(config: dict) -> str:
    return utils.determine_resource_absolute_file(config['playbook'])[0]


def verify_ansible(cluster: KubernetesCluster, config: dict, _plugin_name: Optional[str] = None) -> None:
    _get_absolute_playbook(config)
    if cluster.is_deploying_from_windows():
        raise Exception("Executing of playbooks on Windows deployer is currently not supported")
    # TODO: verify fields types and contents


def apply_ansible(cluster: KubernetesCluster, step: dict) -> None:
    playbook_path = _get_absolute_playbook(step)
    external_vars = step.get('vars', {})
    become = step.get('become', False)
    groups = step.get('groups', [])
    nodes = step.get('nodes', [])

    command = 'ansible-playbook -i ansible-inventory.ini %s' % playbook_path

    if become:
        command += ' -b'

    if groups or nodes:
        common_group = cluster.create_group_from_groups_nodes_names(groups, nodes)
        command += ' --limit %s' % ','.join(common_group.get_nodes_names())

    if external_vars:
        _vars = []
        for k, v in external_vars.items():
            _vars.append('%s=%s' % (k, v))
        command += ' --extra-vars "%s"' % ' '.join(_vars)

    cluster.log.verbose("Running shell \"%s\"" % command)

    result = subprocess.run(command, stdout=sys.stdout, stderr=sys.stderr, shell=True, check=False)
    if result.returncode != 0:
        raise Exception("Failed to apply ansible plugin, see error above")


def apply_helm(cluster: KubernetesCluster, config: dict) -> None:
    chart_path = get_local_chart_path(cluster.log, config)
    process_chart_values(config, chart_path)

    from kubemarine import kubernetes  # pylint: disable=cyclic-import
    local_config_path = kubernetes.fetch_admin_config(cluster)

    with utils.open_external(os.path.join(chart_path, 'Chart.yaml'), 'r') as stream:
        chart_metadata = yaml.safe_load(stream)
        chart_name = chart_metadata["name"]

    cluster.log.debug("Running helm chart %s" % chart_name)

    release = config.get('release', chart_name)
    cluster.log.debug("Deploying release %s" % release)

    namespace = config.get('namespace')
    if not namespace:
        cluster.log.verbose('Namespace configuration is missing, "default" namespace will be used')
        namespace = "default"

    prepare_for_helm_command = f'helm --kubeconfig {local_config_path} -n {namespace} '

    cluster.log.verbose("Check if chart already has been installed")
    # todo probably use single command helm upgrade --install
    command = prepare_for_helm_command + 'list -q'
    helm_existed_releases = execute_subprocess_with_logging(cluster, command, capture_stdout=True)

    if release in helm_existed_releases.splitlines():
        cluster.log.debug("Deployed release %s is found. Upgrading it..." % release)
        deployment_mode = "upgrade"
    else:
        cluster.log.debug("Deployed release %s is not found. Installing it..." % release)
        deployment_mode = "install"

    command = prepare_for_helm_command + f'{deployment_mode} {release} {chart_path} --create-namespace --debug'
    execute_subprocess_with_logging(cluster, command)


def execute_subprocess_with_logging(cluster: KubernetesCluster, command: str, *, capture_stdout: bool = False) -> str:
    """
    :param cluster: KubernetesCluster object
    :param command: command to run through the shell
    :param capture_stdout: If true, log stderr and return stdout. Otherwise, log both stdout and stderr
    :return: child process stdout if `capture_stdout`, or "" otherwise
    """
    process = subprocess.Popen(command, shell=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE if capture_stdout else subprocess.STDOUT,
                               text=True, encoding='utf-8')
    if capture_stdout:
        with process:
            stdout, stderr = process.communicate()
            for line in stderr.splitlines():
                cluster.log.debug(line)
    else:
        stdout_stream = cast(TextIO, process.stdout)
        with stdout_stream:
            for line in iter(stdout_stream.readline, ''):
                cluster.log.debug(line.rstrip('\n'))

        stdout = ""

    retcode = process.wait()
    if retcode:
        raise subprocess.CalledProcessError(retcode, command)

    return stdout


def process_chart_values(config: dict, local_chart_path: str) -> None:
    config_values = config.get("values")
    file_values = None
    config_values_file = config.get("values_file")
    if config_values_file is not None:
        with utils.open_external(config_values_file) as stream:
            file_values = yaml.safe_load(stream)

    if config_values is None and file_values is None:
        return

    chart_values = os.path.join(local_chart_path, 'values.yaml')
    with utils.open_external(chart_values, 'r') as stream:
        merged_values = yaml.safe_load(stream)

    if file_values is not None:
        merged_values = default_merger.merge(merged_values, file_values)
    # Values from 'values' section have priority over values in 'values_file' section
    if config_values is not None:
        merged_values = default_merger.merge(merged_values, config_values)

    utils.dump_file({}, yaml.dump(merged_values), chart_values, dump_location=False)


def get_local_chart_path(logger: log.EnhancedLogger, config: dict) -> str:
    chart_path = config['chart_path']

    is_curl = chart_path[:4] == 'http' and '://' in chart_path[4:8]

    local_chart_folder = "local_chart_folder"
    if os.path.isdir(local_chart_folder):
        shutil.rmtree(local_chart_folder)
    os.makedirs(local_chart_folder)
    if is_curl:
        logger.verbose('Chart download via curl detected')
        destination = os.path.basename(chart_path)

        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        # todo probably add option which will manage if certificate should be verified?
        ctx.verify_mode = ssl.CERT_NONE
        with urllib.request.urlopen(chart_path, context=ctx) as u, \
                open(destination, 'wb') as f:
            shutil.copyfileobj(u, f)

        extension = destination.split('.')[-1]
        if extension == 'zip':
            logger.verbose('Unzip will be used for unpacking')
            with zipfile.ZipFile(destination, 'r') as zf:
                zf.extractall(local_chart_folder)
        else:
            logger.verbose('Tar will be used for unpacking')
            with tarfile.open(destination, "r:gz") as tf:
                tf.extractall(local_chart_folder)
    else:
        logger.debug("Create copy of chart to work with")
        shutil.copytree(chart_path, local_chart_folder, dirs_exist_ok=True)

    # Find all Chart.yaml files in the chart.
    glob_search = os.path.join(local_chart_folder, '**', 'Chart.yaml')
    chart_metadata = glob.glob(glob_search, recursive=True)
    if not chart_metadata:
        raise Exception("Incorrect format of helm chart: Chart.yaml not found")

    # Sort by number of parts in path to find outermost Chart.yaml
    chart_metadata.sort(key=lambda path: len(path.split(os.sep)))
    local_chart_folder = os.path.dirname(chart_metadata[0])
    logger.debug("Detected chart path = %s" % local_chart_folder)

    # Check all nested Chart.yaml are inside chart path
    for i in range(1, len(chart_metadata)):
        if not os.path.commonpath([chart_metadata[i], local_chart_folder]) == local_chart_folder:
            raise Exception(
                f"Incorrect format of helm chart: inner {chart_metadata[i]} is not inside {local_chart_folder} directory.")

    return local_chart_folder


def convert_config(_: KubernetesCluster, config: Union[str, dict]) -> dict:
    return _convert_file(config)


def verify_config(_: KubernetesCluster, config: dict, _plugin_name: Optional[str] = None) -> None:
    _verify_file(config, "Config")


def apply_config(cluster: KubernetesCluster, config: dict) -> None:
    _apply_file(cluster, config, "Config")


def _convert_file(config: Union[str, dict]) -> dict:
    if isinstance(config, str):
        config = {
            'source': config
        }
    return config


def get_source_absolute_pattern(config: dict) -> Tuple[str, bool]:
    abs_dir, is_external = utils.determine_resource_absolute_dir(config['source'])
    basename = os.path.basename(config['source'])
    return os.path.join(abs_dir, basename), is_external


def _verify_file(config: dict, file_type: str) -> None:
    """
        Verifies if the path matching the config 'source' key exists and points to
        existing files.
    """

    # Determite absolute path to templates
    source, _ = get_source_absolute_pattern(config)
    files = glob.glob(source)

    if len(files) == 0:
        raise Exception('Cannot find any %s files matching this '
                        'source value: %s' % (file_type, source))

    for file in files:
        if not os.path.isfile(file):
            raise Exception('%s resource %s is not a file' % (file_type, file))
        # TODO: verify fields types and contents


def _apply_file(cluster: KubernetesCluster, config: dict, file_type: str) -> None:
    """
        Apply yamls as is or
        renders and applies templates that match the config 'source' key.
    """
    from kubemarine.core import defaults  # pylint: disable=cyclic-import

    log = cluster.log
    do_render = config.get('do_render', True)

    source: Union[str, io.StringIO]
    source, is_external = get_source_absolute_pattern(config)
    files = glob.glob(source)

    for file in files:
        cfg_copy = dict(config)
        source_filename = os.path.basename(file)
        source = file
        if do_render:
            # templates usually have '.j2' extension, which we want to remove from resulting filename
            # but we also support usual '.yaml' files without '.j2' extension, in this case we do not want to remove extension
            split_extension = os.path.splitext(source_filename)
            if split_extension[1] == ".j2":
                source_filename = split_extension[0]

            render_vars = {'runtime_vars': cluster.context['runtime_vars'], 'env': kos.Environ()}
            with utils.open_utf8(file, 'r') as template_stream:
                env = defaults.Environment(log, cluster.inventory)
                generated_data = env.from_string(template_stream.read()).render(**render_vars)

            utils.dump_file(cluster, generated_data, source_filename)
            source = io.StringIO(generated_data)
        elif not is_external:
            with utils.open_utf8(file, 'r') as config_stream:
                source = io.StringIO(config_stream.read())

        cfg_copy['source'] = source

        destination_path = cfg_copy.setdefault('destination', '/etc/kubernetes/%s' % source_filename)

        log.debug("Uploading %s..." % file_type)
        log.debug("\tSource: %s" % file)
        log.debug("\tDestination: %s" % destination_path)

        apply_source(cluster, cfg_copy)


def apply_source(cluster: KubernetesCluster, config: dict) -> None:
    """
        Apply resource from 'source' key as is.
    """
    # Set needed settings from config
    apply_required = config.get('apply_required', True)
    use_sudo = config.get('sudo', True)
    destination_groups = config.get('destination_groups', [])
    destination_nodes = config.get('destination_nodes', [])
    apply_groups = config.get('apply_groups', [])
    apply_nodes = config.get('apply_nodes', [])
    source = config['source']
    destination_path = config['destination']
    apply_command = config.get('apply_command', 'kubectl apply -f %s' % destination_path)

    if not destination_groups and not destination_nodes:
        destination_common_group = cluster.nodes['control-plane']
    else:
        destination_common_group = cluster.create_group_from_groups_nodes_names(destination_groups, destination_nodes)

    if not apply_groups and not apply_nodes:
        apply_common_group = cluster.nodes['control-plane'].get_any_member()
    else:
        apply_common_group = cluster.create_group_from_groups_nodes_names(apply_groups, apply_nodes)

    destination_common_group.put(source, destination_path, backup=True, mkdir=True, sudo=use_sudo, compare_hashes=True)

    if apply_required:
        cluster.log.debug("Applying yaml...")
        if use_sudo:
            apply_common_group.sudo(apply_command, hide=False)
        else:
            apply_common_group.run(apply_command, hide=False)
    else:
        cluster.log.debug('Apply is not required')


def procedure_types() -> Dict[str, Dict[str, Callable]]:
    return {
        'template': {
            'convert': convert_template,
            'verify': verify_template,
            'apply': apply_template
        },
        'expect': {
            'convert': convert_expect,
            'apply': apply_expect
        },
        'python': {
            'verify': verify_python,
            'apply': apply_python
        },
        'thirdparty': {
            'verify': verify_thirdparty,
            'apply': apply_thirdparty
        },
        'shell': {
            'convert': convert_shell,
            'verify': verify_shell,
            'apply': apply_shell
        },
        'ansible': {
            'convert': convert_ansible,
            'verify': verify_ansible,
            'apply': apply_ansible
        },
        'helm': {
            'apply': apply_helm
        },
        'config': {
            'convert': convert_config,
            'verify': verify_config,
            'apply': apply_config
        },
    }



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import collections
import io
import itertools
import os
import random
from abc import ABC, abstractmethod
from types import FunctionType
from typing import (
    Callable, Dict, List, Union, Any, TypeVar, Mapping, Iterator, Optional, Iterable, Generic, Set, cast, Sequence
)

from kubemarine.core import utils, log, executor as exe
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.executor import (
    RawExecutor, Token, GenericResult, RunnersResult, HostToResult, Callback, UnexpectedExit,
    RESULT, GroupResult,
)

NodeConfig = Dict[str, Any]
GroupFilter = Union[Callable[[NodeConfig], bool], NodeConfig]

_T = TypeVar('_T', bound=Union['RunnersGroupResult', bool, None])


class GenericGroupResult(GroupResult[RESULT], Mapping[str, RESULT]):

    def __init__(self, cluster: object, results: Mapping[str, RESULT]) -> None:
        super().__init__({host: [result] for host, result in results.items()})
        self.cluster = cluster

    def __getitem__(self, host: str) -> RESULT:
        return self._results[host][0]

    def __len__(self) -> int:
        return len(self._results)

    def __iter__(self) -> Iterator[str]:
        return iter(self._results)

    def get_simple_out(self) -> str:
        return self.get_simple_result().stdout

    def get_simple_result(self) -> RunnersResult:
        if len(self) != 1:
            raise NotImplementedError("Simple result can be returned only for NodeGroupResult consisted of "
                                      "exactly one node, but %s were provided." % list(self.keys()))

        res = list(self.values())[0]
        if not isinstance(res, RunnersResult):
            raise NotImplementedError("It does not make sense to return simple output for result of type %s"
                                      % type(res))

        return res

    def _make_group(self, hosts: Iterable[str]) -> NodeGroup:
        return NodeGroup(hosts, self.cluster)

    def get_group(self) -> NodeGroup:
        """
        Forms and returns a new group from node results
        :return: NodeGroup
        """
        return self._make_group(self.keys())

    def is_any_has_code(self, code: int) -> bool:
        """
        Returns true if some group result has an exit code equal to the given one. Exceptions and other objects in
        results will be ignored.
        :param code: The code with which the result codes will be compared
        :return: Boolean
        """
        for result in self.values():
            if isinstance(result, RunnersResult) and result.exited == code:
                return True
        return False

    def is_any_failed(self) -> bool:
        """
        Returns true if at least one result in the group finished with non-zero code
        :return: Boolean
        """
        return len(self.get_failed_hosts_list()) > 0

    def get_failed_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts whose result finished with non-zero code
        :return: List with hosts
        """
        failed_hosts: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and result.failed:
                failed_hosts.append(host)
        return failed_hosts

    def get_failed_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that either exited with an exception, or the exit code is equals 1
        :return: NodeGroup:
        """
        nodes_list = self.get_failed_hosts_list()
        return self._make_group(nodes_list)

    def get_hosts_list_where_value_in_stdout(self, value: str) -> List[str]:
        """
        Returns a list of hosts that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stdout.
        :return: List with hosts
        """
        hosts_with_stderr_value: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and value in result.stdout:
                hosts_with_stderr_value.append(host)
        return hosts_with_stderr_value

    def get_hosts_list_where_value_in_stderr(self, value: str) -> List[str]:
        """
        Returns a list of hosts that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: List with hosts
        """
        hosts_with_stderr_value: List[str] = []
        for host, result in self.items():
            if isinstance(result, RunnersResult) and value in result.stderr:
                hosts_with_stderr_value.append(host)
        return hosts_with_stderr_value

    def get_nodes_group_where_value_in_stdout(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: NodeGroup
        """
        nodes_list = self.get_hosts_list_where_value_in_stdout(value)
        return self._make_group(nodes_list)

    def get_nodes_group_where_value_in_stderr(self, value: str) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes that contains the given string value in results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: NodeGroup
        """
        nodes_list = self.get_hosts_list_where_value_in_stderr(value)
        return self._make_group(nodes_list)

    def stdout_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stdout.
        :param value: The string value to be found in the nodes results stdout.
        :return: true if string presented
        """
        return len(self.get_hosts_list_where_value_in_stdout(value)) > 0

    def stderr_contains(self, value: str) -> bool:
        """
        Checks for the presence of the given string in all results stderr.
        :param value: The string value to be found in the nodes results stderr.
        :return: true if string presented
        """
        return len(self.get_hosts_list_where_value_in_stderr(value)) > 0


class NodeGroupResult(GenericGroupResult[GenericResult]):
    pass


class RunnersGroupResult(GenericGroupResult[RunnersResult]):
    pass


class CollectorCallback(Callback):
    def __init__(self, cluster: object) -> None:
        self.cluster = cluster
        self.results: Dict[str, List[RunnersResult]] = {}
        """
        List of collected results for each host.
        If all commands are exited for the particular host, the number and order of results
        correspond to the number and order of the queued commands,
        for which the given callback was requested.
        """
        self._result: Optional[RunnersGroupResult] = None

    def accept(self, host: str, token: Token, result: RunnersResult) -> None:
        self.results.setdefault(host, []).append(result)
        self._result = None

    @property
    def result(self) -> RunnersGroupResult:
        """
        Merges the collected `results` into the single RunnersGroupResult instance.
        This can be useful to check merged output of the commands.

        Note that if the result of more than one command is merged,
        one SHOULD NOT use `RunnersResult.exited` and `RunnersResult.command`,
        as the exit code and command of few commands are undefined.

        :return: merged RunnersGroupResult instance.
        """
        if self._result is None:
            self._result = RunnersGroupResult(
                self.cluster,
                {host: RunnersResult.merge(results) for host, results in self.results.items()}
            )

        return self._result


RunResult = Union[RunnersGroupResult, Token]
GROUP_RUN_TYPE = TypeVar('GROUP_RUN_TYPE', bound=RunResult, covariant=True)
GROUP_SELF = TypeVar('GROUP_SELF', bound='AbstractGroup[Union[RunnersGroupResult, Token]]')


class AbstractGroup(Generic[GROUP_RUN_TYPE], ABC):
    def __init__(self, ips: Iterable[Union[str, GROUP_SELF]], cluster: object):
        from kubemarine.core.cluster import KubernetesCluster  # pylint: disable=cyclic-import

        self.cluster = cast(KubernetesCluster, cluster)
        self.nodes: Set[str] = set()
        for ip in ips:
            if isinstance(ip, self.__class__):
                for host in ip.nodes:
                    self.nodes.add(host)
            elif isinstance(ip, str):
                self.nodes.add(ip)
            else:
                raise Exception('Unsupported connection object type')

    @abstractmethod
    def _make_group(self: GROUP_SELF, ips: Iterable[Union[str, GROUP_SELF]]) -> GROUP_SELF:
        pass

    def __eq__(self, other: object) -> bool:
        if self is other:
            return True

        if not isinstance(other, self.__class__):
            return False

        return self.nodes == other.nodes

    def __ne__(self, other: object) -> bool:
        return not self == other

    def run(self, command: str,
            *,
            warn: bool = False, hide: bool = True, pty: bool = False,
            env: Dict[str, str] = None, timeout: int = None,
            callback: Callback = None) -> GROUP_RUN_TYPE:
        """
        Execute `fabric.Connection.run()` for each node in the group.
        Most of kwargs have the same semantics as in `invoke.Runner.run()`
        with probably changed default values, except the `callback`.

        :param command: the shell command to execute
        :param warn: whether to continue on non-zero code, instead of raising a `.GroupException`
        :param hide:
            Do not copy stdout / stderr from remote to local.
            See `invoke.Runner.run()` for details.
        :param pty:
            Enables the actual pseudoterminal. See `invoke.Runner.run()` for details.

            The following properties should be kept in mind:

            - The remote command may behave interactively with `pty=True`, e.g. apply paging, coloring,
              and hang up expecting user input. This may be mitigated by the command-specific flags and env variables.

            - Stdout and stderr are combined. If parsing is required, it should be resistant
              to potentially broken formatting due to unexpected errors / warnings redirected from stderr.
              It may be useful to redirect unnecessary stream to /dev/null

            - The command can be interrupted using SIGINT or ^C, or `timeout`.
              In case of a Ctrl + C, the derived KeyboardInterrupt is thrown containing current output.

        :param env: a dict to update the remote environment.
        :param timeout:
            Close the channel and raise `.GroupException` containing `executor.CommandTimedOut`
            after the specified number of seconds.
            Global timeout for all operations: 2700 seconds.
            **Note:** this does not interrupt the remote command unless `pty=True` is additionally used.
        :param callback:
            Callback to process the result of commands.
            See `executor.Callback.accept()` for details.
        :return:
            Depending on a specific class, real `executor.RunnersResult`,
            or token to obtain the result of deferred command.
        """
        caller: Optional[Dict[str, object]] = None
        if not hide:
            # fetching of the caller info should be at the earliest point
            caller = log.caller_info(self.cluster.log)
        return self._run("run", command, caller,
                         warn=warn, hide=hide, pty=pty,
                         env=env, timeout=timeout, callback=callback)

    def sudo(self, command: str,
             *,
             warn: bool = False, hide: bool = True, pty: bool = False,
             env: Dict[str, str] = None, timeout: int = None,
             callback: Callback = None) -> GROUP_RUN_TYPE:
        """
        Execute `fabric.Connection.sudo()` for each node in the group.

        See `.run()` for details.
        """
        caller: Optional[Dict[str, object]] = None
        if not hide:
            # fetching of the caller info should be at the earliest point
            caller = log.caller_info(self.cluster.log)
        return self._run("sudo", command, caller,
                         warn=warn, hide=hide, pty=pty,
                         env=env, timeout=timeout, callback=callback)

    @abstractmethod
    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]],
             **kwargs: Any) -> GROUP_RUN_TYPE:
        pass

    @abstractmethod
    def get(self, remote_file: str, local_file: str) -> None:
        pass

    def put(self, local_file: Union[io.StringIO, str], remote_file: str,
            *,
            backup: bool = False, sudo: bool = False,
            mkdir: bool = False, immutable: bool = False,
            compare_hashes: bool = False) -> None:
        local_stream = self._prepare_local(local_file, remote_file)
        group_to_upload = self._group_to_upload(local_stream, remote_file, compare_hashes)
        if group_to_upload.is_empty():
            return

        # pylint: disable-next=protected-access
        group_to_upload._put_with_mv(local_stream, remote_file,
                                     backup=backup, sudo=sudo, mkdir=mkdir, immutable=immutable)

    def _prepare_local(self, local_file: Union[io.StringIO, str], remote_file: str) -> Union[bytes, str]:
        if isinstance(local_file, io.StringIO):
            self.cluster.log.verbose("Text is being transferred to remote file \"%s\" on nodes %s"
                                     % (remote_file, list(self.nodes)))
            # This is a W/A to avoid https://github.com/paramiko/paramiko/issues/1133
            # if text contains non-ASCII characters.
            # Use the same encoding as paramiko uses, see paramiko/file.py/BufferedFile.write()

            return local_file.getvalue().encode('utf-8')
        else:
            if not os.path.isfile(local_file):
                raise Exception(f"File {local_file} does not exist")

            self.cluster.log.verbose("Local file \"%s\" is being transferred to remote file \"%s\" on nodes %s"
                                     % (local_file, remote_file, list(self.nodes)))

            return local_file

    def _group_to_upload(self: GROUP_SELF, local_file: Union[bytes, str], remote_file: str,
                         compare_hashes: bool) -> GROUP_SELF:
        if self.is_empty():
            repr_local = '<text>' if isinstance(local_file, bytes) else local_file
            self.cluster.log.verbose(f'No nodes to transfer {repr_local} to remote file {remote_file}')
            return self

        if not compare_hashes:
            return self

        file_size = len(local_file) if isinstance(local_file, bytes) else os.path.getsize(local_file)
        self.cluster.log.verbose(f'File size: %s' % file_size)
        eager_group = self.cluster.make_group(self.nodes)
        local_file_hash = eager_group.get_local_file_sha1(local_file)
        self.cluster.log.verbose('Local file hash: %s' % local_file_hash)
        remote_file_hashes = eager_group.get_remote_file_sha1(remote_file)
        self.cluster.log.verbose('Remote file hashes: %s' % remote_file_hashes)

        hosts_to_upload = []
        for remote_ip, remote_file_hash in remote_file_hashes.items():
            if remote_file_hash != local_file_hash:
                self.cluster.log.verbose('Local and remote hashes does not match on node \'%s\' %s %s'
                                         % (remote_ip, local_file_hash, remote_file_hash))
                hosts_to_upload.append(remote_ip)

        group_to_upload = self._make_group(hosts_to_upload)
        if group_to_upload.is_empty():
            self.cluster.log.verbose('Local and remote hashes are equal on all nodes, no transmission required')

        return group_to_upload

    def _put_with_mv(self, local_stream: Union[bytes, str], remote_file: str,
                     backup: bool, sudo: bool, mkdir: bool, immutable: bool) -> None:

        if sudo:
            self.cluster.log.verbose('A sudoer upload required')

        if backup:
            self.cluster.log.verbose('File \"%s\" backup required' % remote_file)

        if mkdir:
            self.cluster.log.verbose('A parent directory will be created')

        if immutable:
            self.cluster.log.verbose('File \"%s\" immutable set required' % remote_file)

        advanced_move_required = sudo or backup or immutable or mkdir
        temp_filepath = remote_file

        if advanced_move_required:
            # for unknown reason fabric v2 can't put as sudo, and we should use WA via mv
            # also, if we need to backup the file first, then we also have to upload file to tmp first

            temp_filepath = utils.get_remote_tmp_path()
            self.cluster.log.verbose("Uploading to temporary file '%s'..." % temp_filepath)

        self._put(local_stream, temp_filepath)

        if not advanced_move_required:
            return

        self.cluster.log.verbose("Moving temporary file '%s' to '%s'..." % (temp_filepath, remote_file))

        # -Z option is necessary for RHEL family to set SELinux context to default type.
        if sudo:
            mv_command = "sudo chown root:root %s && sudo mv -fZ %s %s" % (temp_filepath, temp_filepath, remote_file)
        else:
            mv_command = "mv -fZ %s %s" % (temp_filepath, remote_file)

        if backup:
            if sudo:
                mv_command = "sudo cp -f %s %s.bak$(sudo ls %s* | sudo wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)
            else:
                mv_command = "cp -f %s %s.bak$(ls %s* | wc -l); %s" \
                             % (remote_file, remote_file, remote_file, mv_command)

        if mkdir:
            file_directory = "/".join(remote_file.split('/')[:-1])
            if sudo:
                mv_command = f"sudo mkdir -p {file_directory}; {mv_command}"
            else:
                mv_command = f"mkdir -p {file_directory}; {mv_command}"

        mv_command = "cmp --silent %s %s || (%s)" % (remote_file, temp_filepath, mv_command)

        if immutable:
            if sudo:
                mv_command = "sudo chattr -i %s; %s; sudo chattr +i %s" % (remote_file, mv_command, remote_file)
            else:
                mv_command = "chattr -i %s; %s; chattr +i %s" % (remote_file, mv_command, remote_file)

        self.sudo(mv_command)

    @abstractmethod
    def _put(self, local_stream: Union[bytes, str], remote_file: str) -> None:
        pass

    def _unsafe_make_runners_result(self, host_results: HostToResult) -> RunnersGroupResult:
        return RunnersGroupResult(self.cluster,
                                  {host: cast(RunnersResult, result) for host, result in host_results.items()})

    def call(self, action: Callable[..., _T], **kwargs: object) -> _T:
        func = cast(FunctionType, action)
        callable_path = "%s.%s" % (func.__module__, func.__name__)
        self.cluster.log.debug("Running %s: " % callable_path)
        result = action(self, **kwargs)
        if isinstance(result, RunnersGroupResult):
            self.cluster.log.debug(result)

        return result

    def call_batch(self: GROUP_SELF, actions: List[Callable[[GROUP_SELF], Any]]) -> None:
        for action in actions:
            self.call(action)

    def get_online_nodes(self: GROUP_SELF, online: bool) -> GROUP_SELF:
        online_hosts = [host for host, node_context in self.cluster.nodes_context.items()
                        if node_context['access']['online'] == online]
        return self._make_group(online_hosts).intersection_group(self)

    def get_accessible_nodes(self: GROUP_SELF) -> GROUP_SELF:
        accessible = [host for host, node_context in self.cluster.nodes_context.items()
                      if node_context['access']['accessible']]
        return self._make_group(accessible).intersection_group(self)

    def get_sudo_nodes(self: GROUP_SELF) -> GROUP_SELF:
        sudo = [host for host, node_context in self.cluster.nodes_context.items()
                if node_context['access']['sudo'] != "No"]
        return self._make_group(sudo).intersection_group(self)

    def get_ordered_members_list(self: GROUP_SELF, apply_filter: GroupFilter = None) -> List[GROUP_SELF]:
        nodes = self.get_ordered_members_configs_list(apply_filter)
        return [self._make_group([node['connect_to']]) for node in nodes]

    def get_ordered_members_configs_list(self, apply_filter: GroupFilter = None) -> List[NodeConfig]:

        if self.is_empty():
            return []

        result = collections.OrderedDict()
        # We have to iterate strictly in order which was defined by user in config-file.
        # By iterative over `previous_inventory` first we ensure that all removed nodes are at the original position,
        # Added nodes will be at the end thus reproducing the procedure enrichment.
        for node in itertools.chain(self.cluster.previous_inventory['nodes'], self.cluster.inventory['nodes']):
            host = node['connect_to']
            # is iterable node from inventory is part of current NodeGroup?
            if host not in self.nodes:
                continue

            if apply_filter is None:
                result[host] = node
                continue

            # apply filters
            suitable = True
            if callable(apply_filter):
                if not apply_filter(node):
                    suitable = False
            else:
                # here intentionally there is no way to filter by values in lists field,
                # for this you need to use custom functions.
                # Current solution implemented in this way because the filtering strategy is
                # unclear - do I need to include when everything matches or is partial matching enough?
                for key, value in apply_filter.items():
                    if node.get(key) is None:
                        suitable = False
                        break
                    if isinstance(value, list):
                        if node[key] not in value:
                            suitable = False
                            break
                    # elif should definitely be here, not if
                    elif node[key] != value:
                        suitable = False
                        break

            # if not filtered
            if suitable:
                result[host] = node

        return list(result.values())

    def get_first_member(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        results = self.get_ordered_members_list(apply_filter)
        if not results:
            raise Exception("Failed to find first group member by the given criteria")
        return results[0]

    def get_any_member(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        member: GROUP_SELF = random.choice(self.get_ordered_members_list(apply_filter))
        self.cluster.log.verbose(f'Selected node {member.get_host()}')
        return member

    def get_member_by_name(self: GROUP_SELF, name: str) -> GROUP_SELF:
        return self.get_first_member({"name": name})

    def new_group(self: GROUP_SELF, apply_filter: GroupFilter = None) -> GROUP_SELF:
        return self._make_group(self.get_ordered_members_list(apply_filter))

    def include_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes.union(group.nodes)
        return self._make_group(ips)

    def exclude_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes - group.nodes
        return self._make_group(ips)

    def intersection_group(self: GROUP_SELF, group: GROUP_SELF) -> GROUP_SELF:
        ips = self.nodes.intersection(group.nodes)
        return self._make_group(ips)

    def get_nodes_names(self) -> List[str]:
        members = self.get_ordered_members_configs_list()
        return [node['name'] for node in members]

    def get_node_name(self) -> str:
        name: str = self.get_config()['name']
        return name

    def get_hosts(self) -> List[str]:
        members = self.get_ordered_members_configs_list()
        return [node['connect_to'] for node in members]

    def get_host(self) -> str:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only host from not a single node")

        return next(iter(self.nodes))

    def get_config(self) -> NodeConfig:
        if len(self.nodes) != 1:
            raise Exception("Cannot get the only node config from not a single node")

        return self.get_ordered_members_configs_list()[0]

    def is_empty(self) -> bool:
        return not self.nodes

    def has_node(self, node_name: str) -> bool:
        return node_name in self.get_nodes_names()

    def having_roles(self: GROUP_SELF, roles: Sequence[str]) -> GROUP_SELF:
        candidate_hosts = set()
        for role in roles:
            candidate_nodes = (self.cluster.previous_nodes, self.cluster.nodes)
            if not any(role in nodes for nodes in candidate_nodes):
                self.cluster.log.verbose(f'Group {role!r} is requested for usage, but this group does not exist.')
                continue

            for nodes in candidate_nodes:
                if role in nodes:
                    candidate_hosts.update(nodes[role].nodes)

        return self.intersection_group(self._make_group(candidate_hosts))

    def nodes_amount(self) -> int:
        """
        Returns the number of nodes within a group
        :return: Integer
        """
        return len(self.nodes)

    def get_nodes_os(self) -> str:
        """
        Returns the detected operating system family for group.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "rhel9",
                 "multiple", "unknown", "unsupported", "<undefined>".
        """
        return self.cluster.get_os_family_for_nodes(self.nodes)

    def is_multi_os(self) -> bool:
        """
        Returns true if same group contains nodes with multiple OS families
        :return: Boolean
        """
        return self.get_nodes_os() == 'multiple'

    def get_subgroup_with_os(self: GROUP_SELF, os_families: Union[str, List[str]]) -> GROUP_SELF:
        """
        Forms and returns a new group from the nodes of the original group that have a specific OS family
        :param os_families: The name of or list of names of required OS families
        :return: NodeGroup
        """
        if not isinstance(os_families, list):
            os_families = [os_families]

        for os_family in os_families:
            if os_family not in ['debian', 'rhel', 'rhel8', 'rhel9']:
                raise Exception('Unsupported OS family provided')
        hosts = []
        for host in self.nodes:
            node_os_family = self.cluster.get_os_family_for_node(host)
            if node_os_family in os_families:
                hosts.append(host)
        return self._make_group(hosts)


class NodeGroup(AbstractGroup[RunnersGroupResult]):
    def _make_group(self: NodeGroup, ips: Iterable[Union[str, NodeGroup]]) -> NodeGroup:
        return NodeGroup(ips, self.cluster)

    def _make_defer(self, executor: RemoteExecutor) -> DeferredGroup:
        return DeferredGroup(self.nodes, self.cluster, executor)

    def new_defer(self) -> DeferredGroup:
        return self.new_executor().group

    def new_executor(self) -> RemoteExecutor:
        return RemoteExecutor(self)

    def get(self, remote_file: str, local_file: str) -> None:
        self._do_exec("get", remote_file, local_file)

    def _put(self, local_stream: Union[bytes, str], remote_file: str) -> None:
        self._do_exec("put", local_stream, remote_file)

    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]],
             **kwargs: Any) -> RunnersGroupResult:
        """
        The method should be called directly from run & sudo without any extra wrappers.
        """
        do_stream = not kwargs['hide']
        if do_stream and len(self.nodes) > 1:
            raise ValueError("Streaming of output is supported only for the single node")

        if do_stream and caller is not None:
            logger = self.cluster.log
            kwargs['out_stream'] = log.LoggerWriter(logger, caller, '\t')
            kwargs['err_stream'] = log.LoggerWriter(logger, caller, '\t')

        results = self._do_exec(do_type, command, **kwargs)
        return self._unsafe_make_runners_result(results)

    def _do_exec(self, do_type: str, *args: object, **kwargs: Any) -> HostToResult:
        callback: Callback = kwargs.pop('callback', None)

        executor = RawExecutor(self.cluster)
        executor.queue(self.get_hosts(), (do_type, args, kwargs), callback=callback)

        excepted = False
        try:
            executor.flush()
        except exe.GroupException:
            excepted = True

        results = {host: results[0]
                   for host, results in executor.get_flat_result().items()}
        if excepted:
            raise GroupResultException(NodeGroupResult(self.cluster, results))

        return results

    def wait_for_reboot(self, initial_boot_history: RunnersGroupResult, timeout: int = None) -> RunnersGroupResult:
        results = self._await_rebooted_nodes(timeout, initial_boot_history=initial_boot_history)
        if any(isinstance(result, BaseException) or result == initial_boot_history.get(host)
               for host, result in results.items()):
            raise GroupResultException(NodeGroupResult(self.cluster, results))

        return self._unsafe_make_runners_result(results)

    def wait_and_get_boot_history(self, timeout: int = None) -> RunnersGroupResult:
        return self.wait_for_reboot(RunnersGroupResult(self.cluster, {}), timeout=timeout)

    def _await_rebooted_nodes(self, timeout: int = None, initial_boot_history: RunnersGroupResult = None) \
            -> HostToResult:

        executor = RawExecutor(self.cluster)
        return executor.wait_for_boot(self.get_hosts(), timeout, initial_boot_history)

    def wait_command_successful(self, command: str,
                                *,
                                retries: int = 15, timeout: int = 5,
                                hide: bool = False, pty: bool = False) -> None:
        logger = self.cluster.log

        def attempt() -> bool:
            result = self.sudo(command, warn=True, hide=hide, pty=pty)
            if hide:
                logger.verbose(result)
                for host in result.get_failed_hosts_list():
                    output = result[host].stdout if pty else result[host].stderr
                    stderr = output.rstrip('\n')
                    if stderr:
                        logger.debug(f"{host}: {stderr}")

            return not result.is_any_failed()

        utils.wait_command_successful(logger, attempt, retries, timeout)

    def wait_commands_successful(self, commands: List[str],
                                 *,
                                 retries: int = 15, timeout: int = 5,
                                 sudo: bool = True, pty: bool = False) -> None:
        if len(self.nodes) != 1:
            raise Exception("Waiting for few commands is currently supported only for single node")

        logger = self.cluster.log
        remained_commands = list(commands)

        def attempt() -> bool:
            defer = self.new_defer()
            for command in remained_commands:
                if sudo:
                    defer.sudo(command, pty=pty)
                else:
                    defer.run(command, pty=pty)

            try:
                defer.flush()
            except GroupException as e:
                results = e.results[self.get_host()]
                for result in results:
                    if isinstance(result, UnexpectedExit):
                        output = result.result.stdout if pty else result.result.stderr
                        logger.debug(output)
                        break
                    if isinstance(result, BaseException):
                        raise

                    del remained_commands[0]

                return False

            return True

        utils.wait_command_successful(logger, attempt, retries, timeout)

    def get_local_file_sha1(self, local_file: Union[bytes, str]) -> str:
        if isinstance(local_file, str):
            return utils.get_local_file_sha1(local_file)
        else:
            return utils.get_stream_sha1(io.BytesIO(local_file))

    def get_remote_file_sha1(self, filename: str) -> Dict[str, Optional[str]]:
        results = self.sudo("openssl sha1 %s" % filename, warn=True)
        return {host: result.stdout.split("= ")[1].strip() if result.stdout else None
                for host, result in results.items()}


class DeferredGroup(AbstractGroup[Token]):
    def __init__(self, ips: Iterable[Union[str, DeferredGroup]], cluster: object, executor: RemoteExecutor):
        super().__init__(ips, cluster)
        self._executor = executor

    @property
    def executor(self) -> RemoteExecutor:
        return self._executor

    def flush(self) -> None:
        self._executor.flush()

    def _make_group(self, ips: Iterable[Union[str, DeferredGroup]]) -> DeferredGroup:
        return DeferredGroup(ips, self.cluster, self._executor)

    def get(self, remote_file: str, local_file: str) -> None:
        self._do_queue("get", remote_file, local_file)

    def _run(self, do_type: str, command: str, caller: Optional[Dict[str, object]], **kwargs: object) -> Token:
        do_stream = not kwargs['hide']
        if do_stream:
            # To support streaming of output with use of RemoteExecutor in deferred mode, it is necessary to:
            # 1) Make sure that no two commands are executed with streaming in parallel to avoid mess in output
            # 2) Do not print output twice if error occurred.
            raise ValueError("Streaming of output is currently not supported in deferred mode")
        return self._do_queue(do_type, command, **kwargs)

    def _put(self, local_stream: Union[bytes, str], remote_file: str) -> None:
        self._do_queue("put", local_stream, remote_file)

    def _do_queue(self, do_type: str, *args: object, **kwargs: Any) -> Token:
        callback: Callback = kwargs.pop('callback', None)
        return self._executor.queue(self.get_hosts(), (do_type, args, kwargs), callback=callback)

    def include_group(self: DeferredGroup, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.include_group(self, group)

    def exclude_group(self, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.exclude_group(self, group)

    def intersection_group(self, group: DeferredGroup) -> DeferredGroup:
        self._check_same_bound_executor(group)
        return AbstractGroup.intersection_group(self, group)

    def _check_same_bound_executor(self, group: DeferredGroup) -> None:
        if self._executor is not group._executor:  # pylint: disable=protected-access
            raise ValueError("Trying to apply set operation on deferred groups bound to different executors")


class RemoteExecutor(RawExecutor):
    def __init__(self, group: NodeGroup, connection_pool: ConnectionPool = None) -> None:
        super().__init__(group.cluster, connection_pool)
        self.group: DeferredGroup = group._make_defer(self)
        self.cluster = group.cluster

    def flush(self) -> None:
        """
        Flushes the connections' queue.
        Throws GroupException in case of any failure.

        :return: grouped tokenized results per connection.
        """
        try:
            super().flush()
        except exe.GroupException as e:
            raise GroupException(self.cluster, e.results) from None


class GroupException(exe.GroupException):
    def __init__(self, cluster: object, results: Mapping[str, Sequence[GenericResult]]):
        super().__init__(results)
        self.cluster = cluster

    def _make_group(self, hosts: Iterable[str]) -> NodeGroup:
        return NodeGroup(hosts, self.cluster)

    def is_any_excepted(self) -> bool:
        """
        Returns true if at least one result in group is an exception

        :return: Boolean
        """
        return len(self.get_excepted_hosts_list()) > 0

    def get_excepted_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts, for which the result is an exception.

        :return: List with hosts
        """
        excepted_hosts: List[str] = []
        for host, results in self.results.items():
            if any(isinstance(result, BaseException) for result in results):
                excepted_hosts.append(host)
        return excepted_hosts

    def get_excepted_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is an exception.

        :return: NodeGroup:
        """
        nodes_list = self.get_excepted_hosts_list()
        return self._make_group(nodes_list)

    def get_exited_hosts_list(self) -> List[str]:
        """
        Returns a list of hosts, for which the result is the completion of all commands.

        :return: List with hosts
        """
        exited_hosts: List[str] = []
        for host, results in self.results.items():
            if all(isinstance(result, RunnersResult) for result in results):
                exited_hosts.append(host)
        return exited_hosts

    def get_exited_nodes_group(self) -> NodeGroup:
        """
        Forms and returns new NodeGroup of nodes, for which the result is the completion of all commands.

        :return: NodeGroup
        """
        nodes_list = self.get_exited_hosts_list()
        return self._make_group(nodes_list)


class GroupResultException(GroupException):
    def __init__(self, result: GenericGroupResult[GenericResult]):
        super().__init__(result.cluster, {host: [res] for host, res in result.items()})
        self.result = result



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from setuptools import setup


def read(path):
    with open(path, 'r', encoding='utf-8') as f:
        return f.read()


VERSION = read("kubemarine/version").strip()

README = read("README.md")
# Special case to refer to anchor within the current README page should be translated to explicit README referring.
README = re.sub(
    r'\[(.*?)]\((#.*?)\)',
    rf'[\1](https://github.com/Netcracker/KubeMarine/blob/{VERSION}/README.md\2)',
    README
)
# Replace all relative links (not starting with http[s]://) to absolute referring to specific version on GitHub
README = re.sub(
    r'\[(.*?)]\((?!https?://)(.*?)\)',
    rf'[\1](https://github.com/Netcracker/KubeMarine/blob/{VERSION}/\2)',
    README
)

# Though deprecated, it seems to be the only way to provide shell scripts.
SCRIPTS=["bin/kubemarine.cmd", "bin/kubemarine"]

setup(
    scripts=SCRIPTS,
    version=VERSION,
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/Netcracker/KubeMarine'
)



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import sys
from typing import List

import yaml

from kubemarine import plugins
from kubemarine.core import static

HELP_DESCRIPTION = """\
Print the supported k8s versions with the respective configurations of third-parties.

How to use:

"""


def make_config() -> dict:
    kubernetes_versions: dict = {}
    for version, compatibility_map in static.KUBERNETES_VERSIONS['compatibility_map'].items():
        kubernetes_version = kubernetes_versions.setdefault(version, {})

        plugins_ = plugins.oob_plugins
        thirdparties_ = ['crictl']

        for software, software_version in compatibility_map.items():
            if software in plugins_:
                kubernetes_version.setdefault('plugins', {}) \
                    .setdefault(software, {})['version'] = software_version

            if software in thirdparties_:
                kubernetes_version.setdefault('thirdparties', {}) \
                    .setdefault(software, {})['version'] = software_version

    return {'kubernetes': kubernetes_versions}


def print_config(cfg: dict, arguments: dict) -> None:
    # pylint: disable=bad-builtin

    format_ = arguments['output']
    if format_ == 'yaml':
        print(yaml.safe_dump(cfg, sort_keys=False), end='')
    elif format_ == 'json':
        print(json.dumps(cfg, indent=4))


def create_context(cli_arguments: List[str] = None) -> dict:
    if cli_arguments is None:
        cli_arguments = sys.argv[1:]

    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     prog='config',
                                     description=HELP_DESCRIPTION)

    parser.add_argument('-o', '--output',
                        choices=['yaml', 'json'], default='yaml',
                        help='output format')

    arguments = vars(parser.parse_args(cli_arguments))
    return {'config_arguments': arguments}


def main(cli_arguments: List[str] = None) -> None:
    arguments = create_context(cli_arguments)['config_arguments']
    cfg = make_config()
    print_config(cfg, arguments)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.plugins import calico, manifest
from kubemarine.plugins.calico import get_calico_manifest_processor, CalicoApiServerManifestProcessor
from kubemarine.plugins.kubernetes_dashboard import DashboardManifestProcessor
from kubemarine.plugins.local_path_provisioner import LocalPathProvisionerManifestProcessor
from kubemarine.plugins.manifest import Identity
from kubemarine.plugins.nginx_ingress import IngressNginxManifestProcessor

MANIFEST_PROCESSOR_PROVIDERS: Dict[Identity, manifest.PROCESSOR_PROVIDER] = {
    Identity("calico"): get_calico_manifest_processor,
    Identity("calico", "apiserver"): CalicoApiServerManifestProcessor,
    Identity("nginx-ingress-controller"): IngressNginxManifestProcessor,
    Identity("kubernetes-dashboard"): DashboardManifestProcessor,
    Identity("local-path-provisioner"): LocalPathProvisionerManifestProcessor,
}


def is_manifest_installed(cluster: KubernetesCluster, manifest_identity: Identity) -> bool:
    """
    Checks if the manifest is (to be) installed during the regular installation procedure.

    :param cluster: KubernetesCluster object
    :param manifest_identity: A pair of (plugin_name, manifest_id) that uniquely identifies the manifest
    :return: `true` is manifest is (to be) installed.
    """
    return get_manifest_processor(cluster, manifest_identity) is not None


def get_manifest_processor(cluster: KubernetesCluster, manifest_identity: Identity) -> Optional[manifest.Processor]:
    """
    Return actual manifest processor if it is (to be) installed during the regular installation procedure.

    :param cluster: KubernetesCluster object
    :param manifest_identity: A pair of (plugin_name, manifest_id) that uniquely identifies the manifest
    :return: actual manifest processor
    """
    if _is_manifest_disabled(cluster.inventory, manifest_identity):
        return None

    config = _get_manifest_installation_step(cluster.inventory, manifest_identity)
    if config is None:
        return None

    return _convert_config_to_manifest_processor(cluster.log, cluster.inventory, config)


def _convert_config_to_manifest_processor(logger: log.VerboseLogger, inventory: dict, config: dict) -> manifest.Processor:
    arguments = config['arguments']
    manifest_identity = Identity(arguments['plugin_name'], arguments.get('manifest_id'))

    return _get_manifest_processor(
        logger, inventory, manifest_identity,
        arguments.get('original_yaml_path'), arguments.get('destination_name'))


def _get_manifest_installation_step(inventory: dict, manifest_identity: Identity) -> Optional[dict]:
    plugin_name = manifest_identity.plugin_name
    items = inventory['plugins'][plugin_name]['installation']['procedures']
    for i, item in enumerate(items):
        if 'python' not in item:
            continue

        config: dict = item['python']
        if config['module'] != "plugins/builtin.py" or config['method'] != "apply_yaml":
            continue

        # presence of arguments and signature of apply_yaml is validated in plugins.verify_python
        arguments = config['arguments']
        declared_name = arguments['plugin_name']
        if declared_name != plugin_name:
            raise Exception(f"Unexpected 'plugin_name={declared_name}' argument "
                            f"in {plugin_name!r} installation step {i}.")

        declared_identity = Identity(plugin_name, arguments.get('manifest_id'))
        if manifest_identity != declared_identity:
            continue

        return config

    return None


def _is_manifest_disabled(inventory: dict, manifest_identity: Identity) -> bool:
    plugin_name = manifest_identity.plugin_name
    if not inventory["plugins"][plugin_name]["install"]:
        return True

    if manifest_identity == Identity("calico", "apiserver") and not calico.is_apiserver_enabled(inventory):
        return True

    return False


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    for manifest_identity in MANIFEST_PROCESSOR_PROVIDERS:
        if _is_manifest_disabled(inventory, manifest_identity):
            continue

        config = _get_manifest_installation_step(inventory, manifest_identity)
        if config is not None:
            processor = _convert_config_to_manifest_processor(cluster.log, inventory, config)
            processor.validate_inventory()
        else:
            cluster.log.warning(f"Invocation of plugins.builtin.apply_yaml for {manifest_identity.repr_id()} "
                                f"is not found for {manifest_identity.plugin_name!r} plugin. "
                                f"Such configuration is obsolete, and support for it may be stopped in future releases.")


def _get_manifest_processor(logger: log.VerboseLogger, inventory: dict, manifest_identity: Identity,
                            original_yaml_path: Optional[str] = None,
                            destination_name: Optional[str] = None) -> manifest.Processor:
    if manifest_identity not in MANIFEST_PROCESSOR_PROVIDERS:
        raise Exception(f"Cannot find processor of {manifest_identity.repr_id()} "
                        f"for {manifest_identity.plugin_name!r} plugin.")

    processor_provider = MANIFEST_PROCESSOR_PROVIDERS[manifest_identity]
    return processor_provider(logger, inventory, original_yaml_path, destination_name)


def apply_yaml(cluster: KubernetesCluster, plugin_name: str,
               manifest_id: Optional[str] = None,
               original_yaml_path: Optional[str] = None,
               destination_name: Optional[str] = None) -> None:
    manifest_identity = Identity(plugin_name, manifest_id)
    # Since the method is called from the inventory, the installation step is certainly present.
    # Though it can still be disabled by other inventory properties.
    if _is_manifest_disabled(cluster.inventory, manifest_identity):
        cluster.log.debug(f"Skip installing of the {manifest_identity.repr_id()} for {plugin_name!r} plugin.")
        return

    processor = _get_manifest_processor(cluster.log, cluster.inventory, manifest_identity,
                                        original_yaml_path, destination_name)

    manifest = processor.enrich()
    processor.apply(cluster, manifest)


def get_namespace_to_necessary_pss_profiles(cluster: KubernetesCluster) -> Dict[str, str]:
    """
    :param cluster: KubernetesCluster object
    :return: minimal required PSS profiles for all installed plugins' namespaces
    """
    result = {}
    for manifest_identity in MANIFEST_PROCESSOR_PROVIDERS:
        # Use _is_manifest_disabled() instead of `not is_manifest_installed()`,
        # because old inventories might still refer to .yaml.j2 templates where PSS was also managed automatically.
        if _is_manifest_disabled(cluster.inventory, manifest_identity):
            continue

        processor = _get_manifest_processor(cluster.log, cluster.inventory, manifest_identity)
        result.update(processor.get_namespace_to_necessary_pss_profiles())

    return result



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module works with sysctl on remote systems.
Using this module you can generate new sysctl configs, install and apply them.
"""

import io
from typing import Dict, Union, Optional, List, cast

from kubemarine.core import utils, inventory
from kubemarine.core.cluster import KubernetesCluster, enrichment, EnrichmentStage
from kubemarine.core.group import NodeGroup, RunnersGroupResult, AbstractGroup, RunResult
from kubemarine.core.yaml_merger import default_merger

ERROR_PID_MAX_NOT_SET = "The 'kernel.pid_max' value is not set for node {node!r}"
ERROR_PID_MAX_EXCEEDS = ("The 'kernel.pid_max' value = {value!r} for node {node!r} "
                         "is greater than the maximum allowable {max!r}")
ERROR_PID_MAX_REQUIRED = ("The 'kernel.pid_max' value = {value!r} for node {node!r} "
                          "is lower than the minimum required for kubelet configuration = {required!r}")
WARN_PID_MAX_LOWER_DEFAULT = ("The 'kernel.pid_max' value = {value!r} for node {node!r} "
                              "is lower than default system value = {default!r}")


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    # Set global default 'kernel.pid_max', and create inventory patches
    _apply_default_pid_max(cluster)

    # Strip obsolete blank values, and strtoint()
    # This should only affect formatting, but not further enrichment.
    _format_inventory(cluster.inventory, [])
    for i, patch in enumerate(cluster.inventory['patches']):
        _format_inventory(patch, ['patches', i])

    # Fill KubernetesCluster.nodes_inventory with sysctl settings
    inventory.patch_inventory(cluster, ['services', 'sysctl'])

    # Convert settings for each node and apply defaults
    _convert_inventory(cluster)
    _apply_common_defaults(cluster)

    # Other verifications
    _verify_pid_max(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['reconfigure'])
def enrich_reconfigure_inventory(cluster: KubernetesCluster) -> None:
    sysctl_config = utils.subdict_yaml(
        cluster.procedure_inventory.get('services', {}), ['sysctl'])

    if sysctl_config:
        default_merger.merge(cluster.inventory.setdefault('services', {}), utils.deepcopy_yaml(sysctl_config))


def _format_inventory(config: dict, path: List[Union[str, int]]) -> None:
    sysctl_config: Dict[str, Union[str, int, dict]] = config.get('services', {}).get('sysctl', {})
    path = path + ['services', 'sysctl']

    for key in sysctl_config:
        _format_value(sysctl_config, key, path)


def _format_value(sysctl_config: Dict[str, Union[str, int, dict]], key: str, path: List[Union[str, int]]) -> None:
    value = sysctl_config[key]
    if isinstance(value, str):
        # Obsolete approach to render values.
        value = value.strip()
        if value == '':
            sysctl_config[key] = value
        else:
            sysctl_config[key] = _strtoint(value, path + [key])


def _convert_inventory(cluster: KubernetesCluster) -> None:
    for node in cluster.nodes['all'].get_ordered_members_list():
        host = node.get_host()
        sysctl_config: Dict[str, Union[str, int, dict]] = cluster.nodes_inventory[host]['services']['sysctl']

        for key in sysctl_config:
            _convert_value(sysctl_config, key)


def _convert_value(sysctl_config: Dict[str, Union[str, int, dict]], key: str) -> None:
    value = sysctl_config[key]
    install: Optional[bool] = None
    if isinstance(value, str):  # value == ''
        value = 0
        install = False

    if isinstance(value, int):
        sysctl_config[key] = value = {
            'value': value,
        }

    if install is not None:
        value['install'] = install


def _strtoint(value: str, path: List[Union[str, int]]) -> int:
    try:
        return utils.strtoint(value)
    except ValueError as e:
        raise ValueError(f"{str(e)} in section {utils.pretty_path(path)}") from None


def _apply_common_defaults(cluster: KubernetesCluster) -> None:
    for node in cluster.nodes['all'].get_ordered_members_list():
        host = node.get_host()
        sysctl_config: Dict[str, dict] = cluster.nodes_inventory[host]['services']['sysctl']

        for key, value in sysctl_config.items():
            if value.get('groups') is None and value.get('nodes') is None:
                value['groups'] = ['control-plane', 'worker', 'balancer']

            value.setdefault('install', True)

            if value.get('nodes') is not None:
                all_nodes_names = cluster.nodes['all'].get_nodes_names()
                unknown_nodes = set(value['nodes']) - set(all_nodes_names)
                if unknown_nodes:
                    # Only warn instead of raising an error to allow remove & add the same node.
                    cluster.log.warning(
                        f"Unknown node names {', '.join(map(repr, unknown_nodes))} "
                        f"provided for kernel parameter {key!r}. ")


def _apply_default_pid_max(cluster: KubernetesCluster) -> None:
    inventory_ = cluster.inventory

    global_pid_max = inventory_['services']['sysctl']['kernel.pid_max']
    if not isinstance(global_pid_max, dict) or 'value' in global_pid_max:
        # If custom value is provided, it should be spread across the specified nodes only
        # unless explicitly overridden in the patches
        return

    global_pid_max.setdefault('value', _get_pid_max(cluster))

    # If kubelet config for some node has different maxPods or podPidsLimit,
    # calculate kernel.pid_max for this node and register it in the default patch.
    nodes_different_value: Dict[int, List[str]] = {}
    for node in cluster.make_group_from_roles(['control-plane', 'worker']).get_ordered_members_list():
        node_pid_max = _get_pid_max(cluster, node)
        if node_pid_max != global_pid_max['value']:
            nodes_different_value.setdefault(node_pid_max, []).append(node.get_node_name())

    for pid_max, nodes in nodes_different_value.items():
        inventory_patch = {
            'nodes': nodes,
            'services': {'sysctl': {
                'kernel.pid_max': pid_max
            }}
        }
        inventory_['patches'].insert(0, inventory_patch)


def _verify_pid_max(cluster: KubernetesCluster) -> None:
    for node in cluster.make_group_from_roles(['control-plane', 'worker']).get_ordered_members_list():
        node_name = node.get_node_name()

        value = get_parameter(cluster, node, 'kernel.pid_max')
        required_pid_max = _get_pid_max(cluster, node)

        if value is None:
            raise Exception(ERROR_PID_MAX_NOT_SET.format(node=node_name))
        if value > 2 ** 22:
            raise Exception(ERROR_PID_MAX_EXCEEDS.format(node=node_name, value=value, max=2 ** 22))
        if value < required_pid_max:
            raise Exception(ERROR_PID_MAX_REQUIRED.format(node=node_name, value=value, required=required_pid_max))
        if value < 32768:
            cluster.log.warning(WARN_PID_MAX_LOWER_DEFAULT.format(node=node_name, value=value, default=32768))


def get_parameter(cluster: KubernetesCluster, node: AbstractGroup[RunResult], key: str) -> Optional[int]:
    host = node.get_host()
    config = cluster.nodes_inventory[host]['services']['sysctl'].get(key, {})

    group = cluster.create_group_from_groups_nodes_names(
        config.get('groups', []), config.get('nodes', []))

    if not config.get('install', False) or not group.has_node(node.get_node_name()):
        return None

    value: int = config['value']
    return value


def make_config(cluster: KubernetesCluster, node: AbstractGroup[RunResult]) -> str:
    """
    Converts parameters from inventory['services']['sysctl'] to a string in the format of sysctl.conf.
    """
    host = node.get_host()
    config = ""
    for key in cluster.nodes_inventory[host]['services']['sysctl']:
        value = get_parameter(cluster, node, key)
        if value is not None:
            config += "%s = %s\n" % (key, value)

    return config


def configure(group: NodeGroup) -> RunnersGroupResult:
    """
    Generates and uploads sysctl configuration to the group.
    The configuration will be placed in sysctl daemon directory.
    """
    cluster: KubernetesCluster = group.cluster
    defer = group.new_defer()

    for node in defer.get_ordered_members_list():
        config = make_config(cluster, node)
        node.sudo('rm -f /etc/sysctl.d/98-*-sysctl.conf')
        utils.dump_file(cluster, config, f'sysctl/98-kubemarine-sysctl_{node.get_node_name()}.conf')
        node.put(io.StringIO(config), '/etc/sysctl.d/98-kubemarine-sysctl.conf', backup=True, sudo=True)

    defer.flush()

    return group.sudo('ls -la /etc/sysctl.d/98-kubemarine-sysctl.conf')


def is_valid(group: NodeGroup) -> bool:
    logger = group.cluster.log

    verify_results = group.sudo('sysctl -a')

    sysctl_valid = True
    for node in group.get_ordered_members_list():
        host = node.get_host()
        result = verify_results[host]
        config = make_config(group.cluster, node)
        for parameter in config.rstrip('\n').split('\n'):
            if parameter not in result.stdout:
                logger.debug(f'Kernel parameter {parameter!r} is not found at {host}')
                sysctl_valid = False

    return sysctl_valid


def reload(group: NodeGroup) -> RunnersGroupResult:
    """
    Reloads sysctl configuration in the specified group.
    """
    return group.sudo('sysctl --system')


def setup_sysctl(group: NodeGroup) -> bool:
    cluster: KubernetesCluster = group.cluster

    if is_valid(group):
        cluster.log.debug("Skipped - all necessary kernel parameters are presented")
        return False

    group.call_batch([
        configure,
        reload,
    ])

    return True


def _get_pid_max(cluster: KubernetesCluster, node: NodeGroup = None) -> int:
    from kubemarine.kubernetes import components  # pylint: disable=cyclic-import

    kubeadm_kubelet = cluster.inventory["services"]["kubeadm_kubelet"]
    max_pods: int = kubeadm_kubelet.get("maxPods", 110)
    pod_pids_limit: int = kubeadm_kubelet.get("podPidsLimit", 4096)

    if node is not None:
        flags = components.get_patched_flags_for_section(cluster, 'kubelet', node)
        if 'maxPods' in flags:
            max_pods = cast(int, flags['maxPods'])
        if 'podPidsLimit' in flags:
            pod_pids_limit = cast(int, flags['podPidsLimit'])

    return max_pods * pod_pids_limit + 2048



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import RunnersGroupResult, NodeGroup
from kubemarine.cri import containerd


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    containerd.enrich_inventory(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def enrich_upgrade_inventory(cluster: KubernetesCluster) -> None:
    containerd.enrich_upgrade_inventory(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def verify_upgrade_inventory(cluster: KubernetesCluster) -> None:
    containerd.verify_upgrade_inventory(cluster)


def install(group: NodeGroup) -> RunnersGroupResult:
    return containerd.install(group)


def configure(group: NodeGroup) -> RunnersGroupResult:
    return containerd.configure(group)


def prune(group: NodeGroup) -> RunnersGroupResult:
    return containerd.prune(group)
